import copy
import json
import os

import numpy as np
import torch
from torch_geometric.data import Data
from torch_geometric.explain.explanation import Explanation
from torch_geometric.graphgym.config import cfg

from explaining_framework.utils.io import read_json, write_json


def _get_explanation(explainer, item):
    explanation = explainer(
        x=item.x,
        edge_index=item.edge_index,
        index=int(item.y),
        target=item.y,
    )
    if not explanation_verification(explanation):
        return None
    else:
        explanation = explanation.to(cfg.accelerator)
        return explanation


def is_empty_graph(data: Data) -> bool:
    return data.x.shape[0] == 0


def get_pred(explainer, explanation):
    pred = explainer.get_prediction(x=explanation.x, edge_index=explanation.edge_index)
    setattr(explanation, "pred", pred)
    data = explanation.to_dict()
    if not data.get("node_mask") is None or not data.get("edge_mask") is None:
        pred_masked = explainer.get_masked_prediction(
            x=explanation.x,
            edge_index=explanation.edge_index,
            node_mask=data.get("node_mask"),
            edge_mask=data.get("edge_mask"),
        )
        setattr(explanation, "pred_exp", pred_masked)


def explanation_verification(exp: Explanation) -> bool:
    is_good = True
    masks = [v for k, v in exp.items() if "_mask" in k and isinstance(v, torch.Tensor)]
    for mask in masks:
        is_nan = mask.isnan().any().item()
        is_inf = mask.isinf().any().item()
        is_ok = exp.validate()
        is_const = mask.max() == mask.min()

        if is_nan or is_inf or not is_ok:
            is_good = False
            return is_good
        else:
            continue
    return is_good


def _save_explanation(exp: Explanation, path: str) -> None:
    data = exp.clone().to_dict()
    for k, v in data.items():
        if isinstance(v, torch.Tensor):
            data[k] = v.clone().detach().cpu().tolist()
    write_json(data, path)


def _load_explanation(path: str) -> Explanation:
    data = read_json(data, path)
    for k, v in data.items():
        if isinstance(v, list):
            if k == "edge_index" or k == "y":
                data[k] = torch.LongTensor(v)
            else:
                data[k] = torch.FloatTensor(v)
    return Explanation.from_dict(data)
